#include "dijkstra.h"

#include "mesh.h"
#include "cluster.h"
extern Mesh init_mesh;
extern PCluster* pcluster;
extern int valid_clusters;
extern int * valid_mapping;

Dijkstra::Dijkstra(int nclusters, int node_size)
{
	nearest_scalar = vector<int>(node_size, -1);
	flooded = vector<bool>(node_size, false);
	propagated = vector<bool>(node_size, false);
	set<int> s;
	cluster_nodes = vector<set<int>>(nclusters, s);

	available_paths.clear();
	closest_dist = vector<float>(node_size, FLT_MAX);
	initialize_neighbors();
}

void Dijkstra::initialize_neighbors()
{
	set<int> s;
	neighbor_nodes = boundary_neighbor_nodes = inner_neighbor_nodes = vector<set<int>>(init_mesh.mesh_node.size(), s);
	for (int i=0; i<init_mesh.mesh_face.size(); i++)
	{
		int v[3];
		for (int j=0; j<3; j++)
		{
			v[j] = init_mesh.mesh_face[i][j];
		}
		
		for (int j=0; j<3; j++)
		{
			for (int k=0; k<3; k++)
			{
				if (v[j] < v[k] ||  (init_mesh.mesh_node[v[j]].boundary_node && init_mesh.mesh_node[v[k]].boundary_node) )
				{
					neighbor_nodes[v[j]].insert(v[k]);
					neighbor_nodes[v[k]].insert(v[j]);

					set<int> neighbor_faces_j = set<int>(init_mesh.node_neighbors[v[j]].begin(), init_mesh.node_neighbors[v[j]].end());
					set<int> neighbor_faces_k = set<int>(init_mesh.node_neighbors[v[k]].begin(), init_mesh.node_neighbors[v[k]].end());
					set<int> shared_neighbors;
					set_intersection(neighbor_faces_j.begin(), neighbor_faces_j.end(), neighbor_faces_k.begin(), neighbor_faces_k.end(), std::inserter(shared_neighbors, shared_neighbors.end()));
					if (shared_neighbors.size() == 1)
					{
						boundary_neighbor_nodes[v[j]].insert(v[k]);
						boundary_neighbor_nodes[v[k]].insert(v[j]);
					}
					else
					{
						int f1 = *shared_neighbors.begin();
						int f2 = *(++shared_neighbors.begin());
						if (pcluster->cluster_belong[f1] != pcluster->cluster_belong[f2])
						{
							boundary_neighbor_nodes[v[j]].insert(v[k]);
							boundary_neighbor_nodes[v[k]].insert(v[j]);
						}
						else
						{
							inner_neighbor_nodes[v[j]].insert(v[k]);
							inner_neighbor_nodes[v[k]].insert(v[j]);
						}
					}
				}
			}
		}
	}
}

void Dijkstra::initialize_boundary_flooding(int cluster_id, vector<pair<int,int>> source_and_scalars)
{
	//get all the nodes in the cluster
	for (set<FaceID>::iterator it = pcluster->face_clusters[valid_mapping[cluster_id]].begin(); it != pcluster->face_clusters[valid_mapping[cluster_id]].end(); it++)
	{
		for (int j = 0; j<3; j++)
		{
			int node_id = init_mesh.mesh_face[*it][j];
			cluster_nodes[cluster_id].insert(node_id);
		}
	}
	//mark as unflooded
	for (set<int>::iterator it = cluster_nodes[cluster_id].begin(); it != cluster_nodes[cluster_id].end(); it++)
	{
		flooded[*it] = false;
		propagated[*it] = false;
		nearest_scalar[*it] = -1;
		closest_dist[*it] = FLT_MAX;
	}

	//set up source node and insert their boundary neighbors into heap
	for (vector<pair<int,int>>::iterator it = source_and_scalars.begin(); it != source_and_scalars.end(); it++)
	{
		flooded[it->first] = true;
		propagated[it->first] = true;
		nearest_scalar[it->first] = it->second;
		closest_dist[it->first] = 0;
		heapify_boundary_neighbors(cluster_id, it->first, true);
		//end inserting into heap
	}
}

void Dijkstra::heapify_boundary_neighbors(int cluster_id, int node_id, bool source)
{
	for (set<int>::iterator n_it = boundary_neighbor_nodes[node_id].begin(); n_it != boundary_neighbor_nodes[node_id].end(); n_it++)
	{
		if (!flooded[*n_it] && cluster_nodes[cluster_id].find(*n_it) != cluster_nodes[cluster_id].end())
		{
			if (!propagated[*n_it])
			{
				PathInfo p;
				p.dist = get_edge_length(node_id, *n_it) + closest_dist[node_id];
				p.prev = node_id;
				p.node_id = *n_it;
				closest_dist[*n_it] = p.dist;
				available_paths.insert(p);
				propagated[*n_it] = true;
			}
			else
			{
				float new_dist = get_edge_length(node_id, *n_it) + closest_dist[node_id];
				if (new_dist < closest_dist[*n_it])
				{
					PathInfo p1;
					p1.dist = closest_dist[*n_it];
					p1.node_id = *n_it;
					p1.prev = -1;
					available_paths.erase(p1);

					PathInfo p2;
					p2.dist = new_dist;
					p2.node_id = *n_it;
					p2.prev = node_id;
					closest_dist[*n_it] = p2.dist;
					available_paths.insert(p2);
				}
			}
		}			
	}
}

double Dijkstra::get_edge_length(int node_i, int node_j)
{
	vec3 v1 = vec3(init_mesh.mesh_node[node_i].x, init_mesh.mesh_node[node_i].y, init_mesh.mesh_node[node_i].z);
	vec3 v2 = vec3(init_mesh.mesh_node[node_j].x, init_mesh.mesh_node[node_j].y, init_mesh.mesh_node[node_j].z);
	return (v1-v2).norm();
}

void Dijkstra::flood_cluster_boundary(int cluster_id)
{
	while(available_paths.size() > 0)
	{
		PathInfo info = *available_paths.begin();
		available_paths.erase(available_paths.begin());
		if (!flooded[info.node_id])
		{	
			flooded[info.node_id] = true;
			nearest_scalar[info.node_id] = nearest_scalar[info.prev];
			heapify_boundary_neighbors(cluster_id, info.node_id, false);
		}
	}
}

void Dijkstra::flood_cluster_inner(int cluster_id)
{
	for (set<int>::iterator it = cluster_nodes[cluster_id].begin(); it != cluster_nodes[cluster_id].end(); it++)
	{
		if (flooded[*it])
			heapify_inner_neighbors(cluster_id, *it, true);
	}

	while(available_paths.size() > 0)
	{
		PathInfo info = *available_paths.begin();
		available_paths.erase(available_paths.begin());
		if (!flooded[info.node_id])
		{	
			flooded[info.node_id] = true;
			nearest_scalar[info.node_id] = nearest_scalar[info.prev];
			heapify_inner_neighbors(cluster_id, info.node_id, false);
		}		
	}
}

void Dijkstra::heapify_inner_neighbors(int cluster_id, int node_id, bool source)
{
	for (set<int>::iterator n_it = inner_neighbor_nodes[node_id].begin(); n_it != inner_neighbor_nodes[node_id].end(); n_it++)
	{
		if (!flooded[*n_it] && cluster_nodes[cluster_id].find(*n_it) != cluster_nodes[cluster_id].end())
		{
			if (!propagated[*n_it])
			{
				PathInfo p;
				p.dist = get_edge_length(node_id, *n_it) + closest_dist[node_id];
				p.prev = node_id;
				p.node_id = *n_it;
				closest_dist[*n_it] = p.dist;
				available_paths.insert(p);
				propagated[*n_it] = true;
			}
			else
			{
				float new_dist = get_edge_length(node_id, *n_it) + closest_dist[node_id];
				if (new_dist < closest_dist[*n_it])
				{
					PathInfo p1;
					p1.dist = closest_dist[*n_it];
					p1.node_id = *n_it;
					p1.prev = -1;
					available_paths.erase(p1);

					PathInfo p2;
					p2.dist = new_dist;
					p2.node_id = *n_it;
					p2.prev = node_id;
					closest_dist[*n_it] = p2.dist;
					available_paths.insert(p2);
				}
			}
		}			
	}
}
